#ifndef _CURSES_H
#define _CURSES_H

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#define LINES 24
#define COLS  79

typedef struct {
    int x, y, dirty;
    char chars[LINES][COLS];
} WINDOW;

static WINDOW stdscr;

static void
clearhome (void)
{
    static char cmd[] = { 27, '[', ';', 'H', 27, '[', '2', 'J' };
    write (1, cmd, sizeof (cmd));
}

static WINDOW *
initscr (void)
{
    memset (stdscr.chars, ' ', LINES*COLS);
    stdscr.dirty = 1;
    return &stdscr;
}

static void
endwin (void)
{
    clearhome ();
}

static void
wclear (WINDOW *w)
{
    memset (w->chars, ' ', LINES*COLS);
    w->dirty = 1;
}

static void
wrefresh (WINDOW *w)
{
    int y;

    if (w->dirty) {
	clearhome ();
	for (y = 1; y < LINES; ++y) {
	    write (1, &w->chars[y][0], COLS);
	    if (y < LINES-1)
	      write (1, "\n", 1);
	}
	w->dirty = 0;
    }
}

static void
wmove (WINDOW *w, int y, int x)
{
    if ((u_int)x >= COLS)
      x = COLS-1;
    if ((u_int)y >= LINES)
      y = LINES-1;
    w->y = y;
    w->x = x;
}

static void
waddch (WINDOW *w, char c)
{
    if (w->chars[w->y][w->x] != c) {
	w->chars[w->y][w->x] = c;
	w->dirty = 1;
    }
    if (++w->x >= COLS) {
	w->x = 0;
	if (++w->y >= LINES)
	  w->y = 0;
    }
}

#endif
