/*
 * User Interface fuer life.
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "curses.h"
#include "RXlib.h"
#include "master.h"

/*
 * RX Variablen...
 */

static RXDisplay *Dpy;
static int        Screen;
static Window     Win = None;
static RXGC       Gc;

/*
 * Curses Variablen
 */

static WINDOW *CWin;

/*
 * sontige Variablen
 */

static char *Bitmap;
static u_int BitmapW, BitmapH;
static int   ScaleW = 1, ScaleH = 1;
static int   DisplayMode = 0;
int          Verzoegerung = 0;
static int   BirthMin = 3, BirthMax = 3;
static int   SurviveMin = 2, SurviveMax = 3;
static int   SeedValue = -1;

extern int Start3D(...);

/*
 * RX11 (De-)Initialisierung
 */
static int
RX11Exit (void)
{
  if (Dpy)
    RXCloseDisplay (Dpy);
  return 0;
}

static int
RX11Init (void)
{
  RXGCValues val;

  if (!Dpy) {
    Dpy = RXOpenDisplay (NULL);
    if (!Dpy) {
      printf ("Kann Verbindung zum RX11 Server nicht herstellen\n");
      printf ("Ist die DISPLAY Variable richtig gesetzt?\n");
      return 1;
    }
    Screen = RXDefaultScreen (Dpy);

    val.foreground = RXBlackPixel (Dpy, Screen);
    val.background = RXWhitePixel (Dpy, Screen);
    Gc = RXCreateGC (Dpy, RXRootWindow (Dpy, Screen),
		     GCForeground | GCBackground, &val);
    return 0;
  }
  return 0;
}

static int
Initialize (void)
{
  static int initialized = 0;

  if (!initialized) {
    CWin = initscr ();
    endwin ();
    initialized = 1;
  }
  return 0;
}

/*
 * Ein paar Hilfsroutinen zum parsen der Eingaben
 */

static int
SkipWS (FILE *fp)
{
    int c;

    while ((c = fgetc (fp)) != EOF && isspace (c))
      ;
    if (c != EOF) {
	ungetc (c, fp);
	return 0;
    }
    return 1;
}

static int
ReadStr (FILE *fp, char **s)
{
    static char str[200];
    char *cp = str;
    int c;

    while ((c = fgetc (fp)) != EOF && isprint (c) && !isspace (c))
      *cp++ = c;
    *cp = 0;
    *s = str;
    return 0;
}

static int
ReadNum (FILE *fp, int *n)
{
    return fscanf (fp, "%d", n) != 1;
}

/*
 * die Kommandos
 */

/*
 * Bitmap einlesen
 */
static int
ReadPixmap (FILE *fp)
{
    int xhot, yhot;
    char *fname;

    if (SkipWS (fp) || ReadStr (fp, &fname))
      return 1;

    if (Bitmap) {
	RXFree (Bitmap);
	Bitmap = NULL;
    }
    Bitmap = RXReadXbmFile (Dpy, fname, &BitmapW, &BitmapH, &xhot, &yhot);
    if (!Bitmap) {
	printf ("Kann xbm Datei '%s' nicht lesen\n", fname);
	return 0;
    }
    printf ("%s: Breite %d, Hoehe %d\n", fname, BitmapW, BitmapH);
    return 0;
}

/*
 * Zufaellige Bitmap erzeugen
 */

static int
RandomBitmap (FILE *fp)
{
    int width, heigth, percent, length, xpos, ypos, theseed;
    char *ptr, *spercent;

    if (SkipWS (fp) || ReadNum (fp, &width) ||
	SkipWS (fp) || ReadNum (fp, &heigth))
      return 1;

    if (width < 2 || heigth < 2) {
	printf ("Das ist mir zu klein.\n");
	return 0;
    }

    if (SkipWS (fp) || ReadStr (fp, &spercent))
      percent = -1;
    else
      percent = atoi (spercent);

    length = ((width+7)/8) * heigth;
    if (Bitmap) {
	RXFree (Bitmap);
	Bitmap = NULL;
    }
    if ((ptr = Bitmap = xmalloc (length)) == NULL) {
	printf ("Kann Bitmap der Groesse %d bytes nicht allokieren\n",
		length);
	return 0;
    }
    memset (Bitmap, '\0', length);

    if (SeedValue < 0)
      theseed = _cputime();
    else
      theseed = SeedValue;

    srand (theseed);

    if (percent <= 0) {
	printf ("Initializing at random.\n");
	for (xpos=0; xpos<length; xpos++)
	  *ptr++ = (char) rand();
    }
    else {
	printf ("Initializing at %d percent\n", percent);
	for (ypos=0; ypos<heigth; ypos++) {
	    for (xpos=1; xpos<width-1; xpos++) {
		if ((rand()%256) * 2 < 5 * percent)
		  (*((ptr) + (ypos) * ((width+7)/8) + ((xpos)>>3)) |= (0x01 << ((xpos) & 0x07)));
	    }
	}
    }
    printf ("random: Breite %d, Hoehe %d, seed %d\n",
	    BitmapW=width, BitmapH=heigth, theseed);
    return 0;
}

/*
 * Skalierung setzen
 */
static int
SetScale (FILE *fp)
{
  int sw, sh;

  if (SkipWS (fp) || ReadNum (fp, &sw) ||
      SkipWS (fp) || ReadNum (fp, &sh))
    return 1;

  if (sw < 1 || sw > 8)
    printf ("Ungueltige X Skalierung, erlaubt sind 1..8\n");
  else
    ScaleW = sw;

  if (sh < 1 || sh > 8)
    printf ("Ungueltige Y Skalierung, erlaubt sind 1..8\n");
  else
    ScaleH = sh;
  return 0;
}

static int
Run (FILE *fp)
{
  int steps;
  int LifeIn3D;

  LifeIn3D = 0;
  if (SkipWS (fp) || ReadNum (fp, &steps))
    return 1;

  if (steps < 0) {
/*    printf ("Was soll ich denn Ihrer Meinung nach bei < 0 Schritten tun?\n");
    return 0;*/
    LifeIn3D = -steps;
  }
  if (!Bitmap) {
    printf ("Es muss erst eine Bitmap mit \"read\" geladen werden.\n");
    return 0;
  }
  if (DisplayMode == 1) {
    /*
     * Graphische Ausgabe
     */
    if (Win != None) {
      /*
       * Fenster vom letzten Mal schliessen
       */
      RXDestroyWindow (Dpy, Win);
      Win = None;
    }
    Win = RXCreateSimpleWindow (Dpy, RXRootWindow (Dpy, Screen), 100, 100,
				BitmapW * ScaleW, BitmapH * ScaleH, 2,
				RXBlackPixel (Dpy, Screen),
				RXWhitePixel (Dpy, Screen));
    RXStoreName (Dpy, Win, "MP-Life");
    if (LifeIn3D != 0)
    {
    	fprintf(stderr,"\nEntering 3D-Mode...\n");
    	fflush(stderr);
    	Start3D(LifeIn3D);
    }
    else
    {
        Calc (Bitmap, BitmapW, BitmapH, steps,
	  BirthMin, BirthMax, SurviveMin, SurviveMax);
    }
  } else if (DisplayMode == 2) {
    /*
     * Textausgabe
     */
    wclear (CWin);
    wrefresh (CWin);
    Calc (Bitmap, BitmapW, BitmapH, steps,
	  BirthMin, BirthMax, SurviveMin, SurviveMax);
    endwin ();
  } else
    Calc (Bitmap, BitmapW, BitmapH, steps,
	  BirthMin, BirthMax, SurviveMin, SurviveMax);
    
  return 0;
}

/*
 * Programm beenden
 */
static int
Exit (FILE *fp)
{
  ProcExit ();
  RX11Exit ();
  return 1;
}

/*
 * Setzen des seed ()
 */

static int
SetSeed (FILE *fp)
{
    char *value;

    if (SkipWS (fp) || ReadStr (fp, &value))
      return 1;

    if (atoi (value) > 0) {
	SeedValue = atoi (value);
	printf ("Seed value set to %d.\n", SeedValue);
    }
    else {
	SeedValue = -1;
	printf ("Seed value set to random.\n");
    }

    return 0;
}

/*
 * Anzeigemodus einstellen
 */
static int
SetDisplay (FILE *fp)
{
  char *mode;

  if (SkipWS (fp) || ReadStr (fp, &mode))
    return 1;

  if (!strcmp (mode, "x11")) {
    if (!RX11Init ())
      DisplayMode = 1;
  } else if (!strcmp (mode, "text")) {
    DisplayMode = 2;
  } else if (!strcmp (mode, "off")) {
    DisplayMode = 0;
  } else
    printf ("Ja was denn nun: x11, txt oder off?\n");
  return 0;
}

/*
 * Verzoegerung setzen
 */
static int
SetDelay (FILE *fp)
{
  int del;

  if (SkipWS (fp) || ReadNum (fp, &del))
    return 1;

  if (del < 0) {
    printf ("Die Veroegerung muss >= 0 sein.\n");
    return 0;
  }
  Verzoegerung = del;
  return 0;
}

/*
 * Ueberlebensregeln setzen
 */
static int
SetRules (FILE *fp)
{
    int bmin, bmax, smin, smax;

    if (SkipWS (fp) || ReadNum (fp, &bmin) ||
	SkipWS (fp) || ReadNum (fp, &bmax) ||
	SkipWS (fp) || ReadNum (fp, &smin) ||
	SkipWS (fp) || ReadNum (fp, &smax))
      return 1;

    if ((u_int)smin > 8 || (u_int)smax > 8)
	printf ("Es muss gelten 0 <= SuriveMin,SurviveMax <= 8\n");
    else {
	SurviveMin = smin;
	SurviveMax = smax;
    }
    if ((u_int)bmin > 8 || (u_int)bmax > 8)
	printf ("Es muss gelten 0 <= BirthMin,BirthMax <= 8\n");
    else {
	BirthMin = bmin;
	BirthMax = bmax;
    }
    return 0;
}

/*
 * Hilfe anzeigen
 */
static int
Help (FILE *fp)
{
    printf ("Verfuegbare Kommandos:\n");
    printf ("\tread <dateiname.xbm>\n");
    printf ("\trun <Schrittzahl>\n");
    printf ("\tscale <X Skalierung> <Y Skalierung>\n");
    printf ("\tdisplay {off|x11|text}\n");
    printf ("\tdelay <Verzoegerung in ms>\n");
    printf ("\trules <Geburt Min> <Geburt Max> "
	    "<Ueberleben Min> <Ueberleben Max>\n");
    printf ("\trandom <Breite> <Hoehe> { <Prozent> | - }\n");
    printf ("\tseed { <value> | - }\n");
    printf ("\texit\n");
    printf ("\thelp\n");
    return 0;
}

static struct {
    char *name;
    int  (*func) (FILE *);
} CommandTab[] = {
    { "read",     ReadPixmap },
    { "exit",     Exit },
    { "help",     Help },
    { "run",      Run },
    { "display",  SetDisplay },
    { "scale",    SetScale },
    { "delay",    SetDelay },
    { "rules",    SetRules },
    { "random",   RandomBitmap },
    { "seed",     SetSeed },
};

/*
 * Warten auf Eingaben
 */

#define Number(x) (sizeof (x)/sizeof (*(x)))

void
MainLoop (FILE *fp)
{
    char *command;
    int i, end = 0;

    Initialize ();

    while (!end) {
	printf ("mp-life> ");
	fflush (stdout);
	if (SkipWS (fp) || ReadStr (fp, &command))
	  break;
	for (i = 0; i < Number (CommandTab); ++i) {
	    if (!strcmp (CommandTab[i].name, command)) {
		if ((*CommandTab[i].func) (fp))
		  end = 1;
		break;
	    }
	}
	if (i >= Number (CommandTab)) {
	    printf ("unbekanntes Kommando: %s\n", command);
	}
    }
}

/*
 * Bitmap darstellen (Aenderungen werden aber erst beim naechsten
 * FlushBitmap () sichtbar).
 */
void
ShowBitmap (char *data, u_int w, u_int h, u_int x, u_int y)
{
  if (DisplayMode == 1) {
    /*
     * X11 Ausgabe
     */
    Pixmap pix;
    char *data2 = data;

    if (ScaleW != 1 || ScaleH != 1) {
      data2 = RXScaleBitmapData (Dpy, data, &w, &h, ScaleW, ScaleH);
      if (!data2) {
	printf ("Kann Bitmap nicht skalieren\n");
	data2 = data;
      }
    }
    RXCopyBitmap (Dpy, Win, Gc, data2, w, h,
		  0, 0, w, h,
		  x * ScaleW, y * ScaleH);
    if (data != data2)
      RXFree (data2);
  } else if (DisplayMode == 2) {
    /*
     * Text Modus
     */
    int i, j, bytes_per_line;
    u_char bit, *byte;

    bytes_per_line = (w+7)/8;

    if (x+w > COLS)
      w = COLS-x;
    if (y+h > LINES)
      h = LINES-y;

    for (i = 0; i < h; ++i) {
      wmove (CWin, y+i, x);

      byte = data + i*bytes_per_line;
      bit  = 0x01;

      for (j = 0; j < w; ++j) {
	if (*byte & bit)
	  waddch (CWin, '*');
	else
	  waddch (CWin, '-');

	if (!(bit <<= 1)) {
	  ++byte;
	  bit = 0x01;
	}
      }
    }
  }
}

/*
 * Ausgabe flushen
 */
void
FlushBitmap (void)
{
  if (DisplayMode == 1)
    RXFlush (Dpy);
  else if (DisplayMode == 2)
    wrefresh (CWin);
}


